import asyncio
import functools
from typing import Any
from urllib.parse import parse_qs
from urllib.parse import urlparse

from twisted.internet.defer import Deferred
from valkey.asyncio import ConnectionPool
from valkey.asyncio import UnixDomainSocketConnection
from valkey.asyncio import Valkey


class PeekWorkerValkeySession:
    def __init__(self, connectionPool: ConnectionPool):
        self.connectionPool = connectionPool
        self.client = None

    async def __aenter__(self):
        self.client = Valkey(connection_pool=self.connectionPool)
        return self

    async def __aexit__(self, excType, excVal, excTb):
        if self.client:
            await self.client.aclose()
            self.client = None

    def __getattr__(self, name: str) -> Any:
        if self.client is None:
            raise RuntimeError(
                "Client not initialized. Use 'async with' context manager."
            )
        return getattr(self.client, name)


class PeekValkeySessionFactory:
    def __init__(self, url: str, maxConnections: int):
        self.connectionPool = ConnectionPool(
            decode_responses=True,
            max_connections=maxConnections,
            retry_on_timeout=True,
            retry_on_error=[ConnectionError, TimeoutError],
            socket_connect_timeout=3600,
            socket_timeout=3600,
            **urlToKwargs(url)
        )

    def __call__(self) -> PeekWorkerValkeySession:
        return PeekWorkerValkeySession(self.connectionPool)

    async def disconnectAsync(self):
        await self.connectionPool.disconnect()
        self.connectionPool = None

    @property
    def createdConnections(self) -> int:
        return len(getattr(self.connectionPool, "_created_connections", []))

    @property
    def availableConnections(self) -> int:
        return len(getattr(self.connectionPool, "_available_connections", []))

    @property
    def inUseConnections(self) -> int:
        return getattr(
            self.connectionPool, "_created_connections_count", 0
        ) - len(getattr(self.connectionPool, "_available_connections", []))

    @property
    def maxConnections(self) -> int:
        return self.connectionPool.max_connections


def convertAsyncioMethodToDeferred(method):
    @functools.wraps(method)
    def wrapper(*args, **kwargs) -> Deferred:
        return Deferred.fromFuture(asyncio.create_task(method(*args, **kwargs)))

    return wrapper


def urlToKwargs(url: str) -> dict:
    parsed = urlparse(url)
    kwargs = {}

    if parsed.scheme == "unix":
        kwargs["connection_class"] = UnixDomainSocketConnection
        kwargs["path"] = parsed.path

        if parsed.query:
            query_params = parse_qs(parsed.query)
            if "db" in query_params:
                kwargs["db"] = int(query_params["db"][0])

    elif parsed.scheme in ["valkey", "redis"]:
        kwargs["host"] = parsed.hostname
        kwargs["port"] = parsed.port or 6379
        kwargs["db"] = (
            int(parsed.path.lstrip("/"))
            if parsed.path and parsed.path != "/"
            else 0
        )

        if parsed.password:
            kwargs["password"] = parsed.password
        if parsed.username:
            kwargs["username"] = parsed.username

    return kwargs



#!/usr/bin/env python
from twisted.internet import asyncioreactor
from twisted.internet.defer import inlineCallbacks

# CRITICAL: Install asyncioreactor BEFORE importing any other twisted modules
asyncioreactor.install()

import logging

from reactivex import operators
from twisted.internet import reactor
from vortex.DeferUtil import vortexLogFailure
from vortex.VortexFactory import VortexFactory

from peek_platform.file_config.PeekFileConfigWorkerMixin import (
    PeekFileConfigWorkerMixin,
)
from peek_platform.platform_init.init_platform import InitPlatform
from peek_platform.util.LogUtil import setupPeekLogger
from peek_plugin_base.PeekVortexUtil import peekServerName
from peek_plugin_base.PeekVortexUtil import peekWorkerName
from peek_worker_service.peek_worker_manager import PeekWorkerManager
from peek_worker_service.peek_worker_manager import WorkerConfig


logger = logging.getLogger(__name__)


def main():
    # Setup the logger here, otherwise subprocsseses will call it
    setupPeekLogger(peekWorkerName)
    # defer.setDebugging(True)
    # sys.argv.remove(DEBUG_ARG)
    # import pydevd
    # pydevd.settrace(suspend=False)

    platformIniter = InitPlatform(peekWorkerName)

    platformIniter.setupPlatform()

    from peek_platform import PeekPlatformConfig

    config = PeekPlatformConfig.config
    assert isinstance(
        config, PeekFileConfigWorkerMixin
    ), "config is not of type PeekFileConfigWorkerMixin"

    workerConfig = WorkerConfig(
        url=config.taskValkeyUrl,
        numWorkers=config.taskWorkerCount,
        maxWorkerTasksCompleted=config.taskReplaceWorkerAfterTasksCompleted,
        maxWorkerAgeSeconds=config.taskReplaceWorkerAfterSeconds,
        maxMemoryMb=config.taskReplaceWorkerAfterMemUsage,
        enabledPlugins=PeekPlatformConfig.config.pluginsEnabled,
    )
    workerProcess = PeekWorkerManager(workerConfig)

    @inlineCallbacks
    def startTasks():
        yield workerProcess.start()

        reactor.addSystemEventTrigger(
            "before", "shutdown", lambda: workerProcess.shutdown()
        )

    @inlineCallbacks
    def restart():
        yield workerProcess.shutdown()
        yield platformIniter.stopAndShutdownPluginsAndVortex()
        PeekPlatformConfig.peekSwInstallManager.restartProcess()

    (
        VortexFactory.subscribeToVortexStatusChange(peekServerName)
        .pipe(operators.filter(lambda online: online == False))
        .subscribe(on_next=lambda _: reactor.callLater(0, restart))
    )

    # Start Update Handler,
    # Add both, The peek client might fail to connect, and if it does, the payload
    # sent from the peekSwUpdater will be queued and sent when it does connect.

    # Software update check is not a thing any more
    # d.addErrback(vortexLogFailure, logger, consumeError=True)
    # d.addCallback(lambda _: peekSwVersionPollHandler.start())

    # Start client main data observer, this is not used by the plugins
    # (Initialised now, not as a callback)

    # Load all Plugins
    d = platformIniter.connectVortexClient()
    d.addCallback(lambda _: platformIniter.loadAndStartupPlugins())
    d.addCallback(lambda _: startTasks())

    def startedSuccessfully(_):
        logger.info(
            "Peek Worker is running, version=%s",
            PeekPlatformConfig.config.platformVersion,
        )
        return _

    d.addErrback(vortexLogFailure, logger, consumeError=False)
    d.addErrback(lambda _: restart())
    d.addCallback(startedSuccessfully)

    reactor.addSystemEventTrigger(
        "before", "shutdown", platformIniter.stopAndShutdownPluginsAndVortex
    )

    reactor.run()


if __name__ == "__main__":
    main()



"""
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
 *
"""

import logging

from peek_platform.file_config.PeekFileConfigABC import PeekFileConfigABC
from peek_platform.file_config.PeekFileConfigDataExchangeClientMixin import (
    PeekFileConfigDataExchangeClientMixin,
)
from peek_platform.file_config.PeekFileConfigOsMixin import (
    PeekFileConfigOsMixin,
)
from peek_platform.file_config.PeekFileConfigPlatformMixin import (
    PeekFileConfigPlatformMixin,
)
from peek_platform.file_config.PeekFileConfigSqlAlchemyMixin import (
    PeekFileConfigSqlAlchemyMixin,
)
from peek_platform.file_config.PeekFileConfigWorkerMixin import (
    PeekFileConfigWorkerMixin,
)

logger = logging.getLogger(__name__)


class PeekWorkerConfig(
    PeekFileConfigABC,
    PeekFileConfigPlatformMixin,
    PeekFileConfigOsMixin,
    PeekFileConfigSqlAlchemyMixin,
    PeekFileConfigWorkerMixin,
):
    """
    This class creates a basic worker configuration
    """

    def __init__(self):
        super().__init__()
        self.dataExchange = PeekFileConfigDataExchangeClientMixin(self)



from twisted.internet import asyncioreactor
from twisted.internet.error import ReactorAlreadyInstalledError

try:
    # CRITICAL: Install asyncioreactor BEFORE importing any other twisted modules
    asyncioreactor.install()
except ReactorAlreadyInstalledError:
    # This is a subprocess, so this is expected. Continue running.
    pass

import logging
import os
import signal
import sys
import uuid
from datetime import datetime
from datetime import timezone
from importlib.util import find_spec
from typing import Optional

from setproctitle import setproctitle
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from vortex import Tuple as TupleMod
from vortex.DeferUtil import deferToThreadWrapWithLogger

from peek_platform.util.LogUtil import setupPeekLogger
from peek_plugin_base.PeekVortexUtil import peekWorkerName
from peek_worker_service.peek_worker_task import PeekWorkerTask

# This is a subprocess, so __main__ doensn't work
logger = logging.getLogger("peek_plugin_worker.peek_worker_process")


class PeekWorkerProcess:
    def __init__(
        self,
        TUPLE_TYPES,
        TUPLE_TYPES_BY_NAME,
        TUPLE_TYPES_BY_SHORT_NAME,
        url: str,
        enabledPlugins: list[str],
    ):
        from peek_worker_service.peek_worker_process_queue import (
            PeekWorkerProcessQueue,
        )

        self._TUPLE_TYPES = TUPLE_TYPES
        self._TUPLE_TYPES_BY_NAME = TUPLE_TYPES_BY_NAME
        self._TUPLE_TYPES_BY_SHORT_NAME = TUPLE_TYPES_BY_SHORT_NAME

        self.workerId: str = str(uuid.uuid4())[:8]
        self.url = url
        self.processId: Optional[int] = None
        self.workerQueue: Optional["PeekWorkerProcessQueue"] = None
        self.running: bool = False
        self.shutdownRequested: bool = False
        self._tasksProcessed: int = 0

        self._enabledPluginsSet = set(enabledPlugins)

    @property
    def totalTasksProcessed(self) -> int:
        return self._tasksProcessed

    @property
    def isAlive(self) -> bool:
        return (
            self.running
            and not self.shutdownRequested
            and self.workerQueue
            and self.workerQueue.isActive
        )

    def start(self):
        """Start

        This is called by the subprocess
        """
        self.running = True
        self.processId = os.getpid()
        signal.signal(signal.SIGTERM, self.multiProcessShutdown)
        signal.signal(signal.SIGINT, self.multiProcessShutdown)

        TupleMod.TUPLE_TYPES = self._TUPLE_TYPES
        TupleMod.TUPLE_TYPES_BY_NAME = self._TUPLE_TYPES_BY_NAME
        TupleMod.TUPLE_TYPES_BY_SHORT_NAME = self._TUPLE_TYPES_BY_SHORT_NAME

        logger.info(f"started (PID: {self.processId})")
        setupPeekLogger(peekWorkerName, self.workerId, logToStdout=False)
        setproctitle(f"{peekWorkerName} {self.workerId}")

        logger.debug(f"starting reactor")
        # noinspection PyUnresolvedReferences
        reactor.callLater(0, self._startInReactor)
        # noinspection PyUnresolvedReferences
        reactor.run()

    def multiProcessShutdown(self, signum: int, frame):
        logger.info(
            f"received signal {signum}," f" initiating graceful shutdown"
        )
        self.shutdownRequested = True
        self.running = False

        # noinspection PyUnresolvedReferences
        reactor.callFromThread(self._shutdownInReactor)

    @inlineCallbacks
    def _startInReactor(self):

        from peek_worker_service.peek_worker_process_queue import (
            PeekWorkerProcessQueue,
        )

        self.workerQueue = PeekWorkerProcessQueue(self.url)
        yield self.workerQueue.start()

        # noinspection PyUnresolvedReferences
        reactor.callLater(0, self._checkAndProcessLoop)

    @inlineCallbacks
    def _shutdownInReactor(self):
        logger.info(f"shutting down gracefully")

        try:
            if self.workerQueue:
                yield self.workerQueue.shutdown()
        except Exception as e:
            logger.error(f"error closing worker queue: {e}")

        # noinspection PyUnresolvedReferences
        if reactor.running:
            # noinspection PyUnresolvedReferences
            reactor.stop()

    @inlineCallbacks
    def _checkAndProcessLoop(self):
        nextTaskDeferred = None
        while self.isAlive:

            try:
                if nextTaskDeferred:
                    taskDoc, taskArgs = yield nextTaskDeferred
                    nextTaskDeferred = None
                else:
                    result = yield self.workerQueue.waitForTask(self.workerId)
                    if not result:
                        continue
                    taskDoc, taskArgs = result

                if not taskDoc:
                    if self.shutdownRequested:
                        logger.info(f"exiting due to shutdown request")
                        break
                    logger.debug(f"waitForTask got None")
                    continue

                from peek_worker_service.peek_worker_task_state import TaskState

                if taskDoc.get("status") == TaskState.CANCELLED.value:
                    logger.info(f"skipped cancelled task {taskDoc['_id'][:8]}")
                    continue

                # Only prefetch next task if not shutting down
                if not self.shutdownRequested:
                    nextTaskDeferred = self.workerQueue.waitForTask(
                        self.workerId
                    )

                yield self._processTask(taskDoc, taskArgs)

            except Exception as e:
                logger.error(f"error in task loop: {e}")

            if not self.isAlive:
                logger.debug(f"exiting task loop")
                break

        logger.debug(f"task processing loop completed")

    @inlineCallbacks
    def _processTask(self, taskDoc: dict, taskArgs):
        taskId: str = taskDoc["_id"]
        taskMethodName: str = taskArgs.taskMethod
        args: list = taskArgs.args
        kwargs: dict = taskArgs.kwargs

        startTime = datetime.now(timezone.utc)

        logger.info(f"started task {taskId[:8]} ({taskMethodName})")

        from peek_worker_service.peek_worker_tuples import (
            PeekWorkerTaskResultTuple,
        )

        resultTuple = PeekWorkerTaskResultTuple()
        resultTuple.taskId = taskId
        resultTuple.completedAt = None
        resultTuple.startedAt = None
        resultTuple.finishedAt = None
        resultTuple.dequeuedAt = None
        resultTuple.cancelledAt = None
        resultTuple.cancellationReason = None

        try:
            pluginName = taskMethodName.split(".")[0]
            if (
                not pluginName.startswith("peek_core")
                and pluginName not in self._enabledPluginsSet
            ):
                raise Exception(
                    f"taskMethodWrapper '{taskMethodName}' is for a plugin"
                    f" not enabled in config.json"
                )

            result = yield self._runTaskMethod(taskMethodName, args, kwargs)

            resultTuple.result = result
            resultTuple.error = None
            yield self.workerQueue.queueResult(taskId, resultTuple)
            logger.info(
                f"completed task {taskId[:8]},"
                f" in {datetime.now(timezone.utc) - startTime}"
            )
            self._tasksProcessed = self._tasksProcessed + 1

        except Exception as e:
            errorMsg = str(e)
            logger.exception(f"failed task {taskId[:8]}: {errorMsg}")

            resultTuple.result = None
            resultTuple.error = errorMsg
            yield self.workerQueue.queueResult(taskId, resultTuple)

    @deferToThreadWrapWithLogger(logger)
    def _runTaskMethod(self, taskMethodName: str, args, kwargs):

        taskMethodWrapper = self._loadTaskMethodBlocking(taskMethodName)
        return taskMethodWrapper(*args, **kwargs)

    def _loadTaskMethodBlocking(self, taskMethodName: str) -> PeekWorkerTask:

        modName, functionName = taskMethodName.rsplit(".", 1)

        if modName in sys.modules:
            module = sys.modules[modName]
        else:
            modSpec = find_spec(modName)
            if not modSpec:
                raise Exception(
                    f"Failed to find module {modName}, is the python package installed?"
                )
            module = modSpec.loader.load_module()

        taskMethodWrapper = getattr(module, functionName)

        assert isinstance(taskMethodWrapper, PeekWorkerTask), (
            "taskMethodWrapper %s is not a PeekWorkerTask" % taskMethodName
        )
        return taskMethodWrapper



import logging
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Dict

from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from twisted.internet.task import LoopingCall
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload

from peek_plugin_base.LoopingCallUtil import peekCatchErrbackWithLogger
from peek_worker_service.peek_worker_task_state import TaskState
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskArgsTuple
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskResultTuple
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskStartTuple
from peek_worker_service.peek_worker_valkey_session import (
    PeekValkeySessionFactory,
)
from peek_worker_service.peek_worker_valkey_session import (
    convertAsyncioMethodToDeferred,
)

logger = logging.getLogger(__name__)


class PeekWorkerRequestQueue:
    LOOPING_CALL_PERIOD = 3600  # 1 hour in seconds
    MONITORING_PERIOD = 300  # 5 minutes in seconds
    MAX_CONCURRENT_TASK_REQUEST_CONNECTIONS = 10
    MAX_CONCURRENT_TASK_RESULT_CONNECTIONS = 2

    def __init__(self, url: str = "valkey://localhost:6379/0"):
        self.url = url

        self._poolRequestSemphore = defer.DeferredSemaphore(
            tokens=self.MAX_CONCURRENT_TASK_REQUEST_CONNECTIONS
        )

        self._valkeySessionCreator: PeekValkeySessionFactory = None

        self.taskQueueKey = "task_queue"
        self.taskArgsKey = "task_args"
        self.resultQueueKey = "result_queue"
        self.taskTimingKey = "task_timing"
        self.pendingTasksKey = "pending_tasks"
        self._isAlive = False
        self._runningTasksByTaskId: Dict[str, "PeekTaskResult"] = {}
        self._loopRunning = False

        # Initialize looping calls
        self._cleanupLoopingCall = LoopingCall(
            peekCatchErrbackWithLogger(logger)(self._removeOldTasksWrapper)
        )
        self._monitoringLoopingCall = LoopingCall(
            peekCatchErrbackWithLogger(logger)(self._monitorValkeyStatus)
        )

    def serialise(self, tuple_) -> str:
        # noinspection PyProtectedMember
        return Payload(tuples=[tuple_])._toJson()

    # noinspection SpellCheckingInspection
    def deserialise(self, jsonStr: str):
        # noinspection PyProtectedMember
        return Payload()._fromJson(jsonStr).tuples[0]

    # noinspection SpellCheckingInspection
    @deferToThreadWrapWithLogger(logger)
    def deserialiseInThread(self, jsonStr: str):
        return self.deserialise(jsonStr)

    @property
    def isActive(self) -> bool:
        return self._isAlive and self._valkeySessionCreator is not None

    @convertAsyncioMethodToDeferred
    async def start(self):
        self._isAlive = True
        self._valkeySessionCreator = PeekValkeySessionFactory(
            url=self.url,
            maxConnections=(
                self.MAX_CONCURRENT_TASK_RESULT_CONNECTIONS
                + self.MAX_CONCURRENT_TASK_REQUEST_CONNECTIONS
                + 10  # Add some buffer
            ),
        )

        # Test connection
        async with self._valkeySessionCreator() as client:
            await client.ping()
            logger.info(f"Successfully connected to Valkey at {self.url}")

        # Clear queues on startup
        await self._cleanupCollectionsAsync()

        # Start looping calls
        d1 = self._cleanupLoopingCall.start(self.LOOPING_CALL_PERIOD, now=True)
        d1.addErrback(vortexLogFailure, logger, consumeError=True)

        d2 = self._monitoringLoopingCall.start(self.MONITORING_PERIOD, now=True)
        d2.addErrback(vortexLogFailure, logger, consumeError=True)

    def shutdown(self):
        self._isAlive = False

        if self._cleanupLoopingCall is not None:
            self._cleanupLoopingCall.stop()
            self._cleanupLoopingCall = None

        if self._monitoringLoopingCall is not None:
            self._monitoringLoopingCall.stop()
            self._monitoringLoopingCall = None

    @convertAsyncioMethodToDeferred
    async def _monitorValkeyStatus(self):
        if not self._valkeySessionCreator:
            return

        try:
            async with self._valkeySessionCreator() as client:
                info = await client.info()

                # Log interesting connection and memory stats
                connectedClients = info.get("connected_clients", "N/A")
                usedMemory = info.get("used_memory_human", "N/A")
                totalConnectionsReceived = info.get(
                    "total_connections_received", "N/A"
                )
                totalCommandsProcessed = info.get(
                    "total_commands_processed", "N/A"
                )
                keyspaceHits = info.get("keyspace_hits", "N/A")
                keyspaceMisses = info.get("keyspace_misses", "N/A")

                logger.info(
                    f"Valkey Status - Connected clients: {connectedClients}, "
                    f"Memory used: {usedMemory}, "
                    f"Total connections: {totalConnectionsReceived}, "
                    f"Commands processed: {totalCommandsProcessed}, "
                    f"Keyspace hits: {keyspaceHits}, misses: {keyspaceMisses}"
                )

                # Log pool stats - use async pool properties
                inUseConnections = self._valkeySessionCreator.inUseConnections
                waitingConnections = max(
                    0,
                    inUseConnections
                    - self._valkeySessionCreator.maxConnections,
                )
                logger.info(
                    f"Connection Pool - Max: "
                    f"{self._valkeySessionCreator.maxConnections}, "
                    f"Created: {self._valkeySessionCreator.createdConnections}, "
                    f"Available: {self._valkeySessionCreator.availableConnections}, "
                    f"In Use: {inUseConnections}, "
                    f"Waiting: {waitingConnections}"
                )

        except Exception as e:
            logger.error(f"Error monitoring Valkey status: {e}")

    def queueTask(
        self,
        taskMethod: str,
        requestArgs: PeekWorkerTaskStartTuple,
        taskArgs: PeekWorkerTaskArgsTuple,
        timeout: float,
    ) -> "TaskResult":
        from peek_worker_service.peek_worker_task_result import PeekTaskResult

        taskId = requestArgs.taskId
        taskResult = PeekTaskResult(self, taskId, timeout)

        # Store the task result in running tasks
        self._runningTasksByTaskId[taskId] = taskResult

        # Queue the task
        queueD = self._poolRequestSemphore.run(
            self._queueTask, taskMethod, requestArgs, taskArgs
        )
        queueD.addErrback(
            lambda f: logger.error(f"Error queueing task: {f.value}")
        )
        reactor.callLater(0, self._resultLoopSingleton)

        return taskResult

    @convertAsyncioMethodToDeferred
    async def _queueTask(
        self,
        taskMethod: str,
        requestArgs: PeekWorkerTaskStartTuple,
        taskArgs: PeekWorkerTaskArgsTuple,
    ):
        taskId = requestArgs.taskId
        logger.debug(f"Queuing task {taskId[:8]} ({taskMethod})")

        async with self._valkeySessionCreator() as client:
            # Store task arguments separately
            taskArgsKey = f"{self.taskArgsKey}:{taskId}"
            serializedArgs = self.serialise(taskArgs)
            await client.set(taskArgsKey, serializedArgs)

            # Store task metadata without arguments
            taskDoc = {
                "_id": taskId,
                "status": requestArgs.status,
                "created_at": requestArgs.createdAt.isoformat(),
                "queued_at": (
                    requestArgs.queuedAt.isoformat()
                    if requestArgs.queuedAt
                    else None
                ),
                "worker_id": requestArgs.workerId,
                "started_at": (
                    requestArgs.startedAt.isoformat()
                    if requestArgs.startedAt
                    else None
                ),
                "dequeued_at": (
                    requestArgs.dequeuedAt.isoformat()
                    if requestArgs.dequeuedAt
                    else None
                ),
                "finished_at": (
                    requestArgs.finishedAt.isoformat()
                    if requestArgs.finishedAt
                    else None
                ),
                "cancelled_at": (
                    requestArgs.cancelledAt.isoformat()
                    if requestArgs.cancelledAt
                    else None
                ),
                "cancellation_reason": requestArgs.cancellationReason,
            }

            # Store task data using vortex serialization
            taskKey = f"{self.taskQueueKey}:{taskId}"
            taskDocSerialized = self.serialise(taskDoc)
            await client.set(taskKey, taskDocSerialized)

            # Add to pending tasks list for workers to block on
            await client.lpush(self.pendingTasksKey, taskId)

        logger.debug(f"Task {taskId[:8]} successfully queued")

    def _resultLoopSingleton(self):
        """Singleton wrapper for the result loop"""
        if self._loopRunning:
            return

        if not self._runningTasksByTaskId:
            return

        def setNotRunning(*args):
            self._loopRunning = False

        self._loopRunning = True
        d = self._resultLoop()
        d.addErrback(lambda f: logger.error(f"Error in result loop: {f.value}"))
        d.addBoth(setNotRunning)

    @convertAsyncioMethodToDeferred
    async def _resultLoop(self):
        """Combined result loop method"""

        async with self._valkeySessionCreator() as client:
            while self.isActive and self._runningTasksByTaskId:
                # Check for timed out tasks
                timedOutTasks = []
                for taskId, taskResult in self._runningTasksByTaskId.items():
                    if taskResult.hasTimedOut:
                        timedOutTasks.append(taskId)

                # Handle timed out tasks
                for taskId in timedOutTasks:
                    logger.info(f"Task {taskId[:8]} timed out")
                    taskResult = self._runningTasksByTaskId.pop(taskId)
                    await self._markTaskAsCancelledAsync(
                        taskId,
                        f"Timeout waiting for result after {taskResult._timeout}s",
                    )
                    taskResult.errback(
                        TimeoutError(
                            f"Timeout waiting for result after {taskResult._timeout}s"
                        )
                    )

                if not self._runningTasksByTaskId:
                    return

                while True:
                    result = await client.blpop(self.resultQueueKey, timeout=5)
                    if not result:
                        break

                    reactor.callLater(0, self._handleResult, result)

    @inlineCallbacks
    def _handleResult(self, result):
        """Handle a result from the queue"""
        assert result, "_handleResult: Result must not be None"

        # Deserialize result
        resultTuple = yield self.deserialiseInThread(result[1])
        taskId = resultTuple.taskId

        # Get the running task
        if taskId not in self._runningTasksByTaskId:
            logger.warning(f"Received result for unknown task {taskId[:8]}")
            return

        taskResult = self._runningTasksByTaskId.pop(taskId)

        if resultTuple.startedAt:
            took = datetime.now(timezone.utc) - resultTuple.startedAt
        else:
            took = "[startedAt is None]"

        logger.info(f"Task {taskId[:8]} result received, took {took}")

        # Resolve the deferred
        if taskResult.called:
            logger.debug(f"Task {taskId[:8]} deferred already called")
        elif resultTuple.error:
            taskResult.errback(Exception(resultTuple.error))
        else:
            taskResult.callback(resultTuple.result)

    async def _markTaskAsCancelledAsync(self, taskId: str, reason: str = None):
        now = datetime.now(timezone.utc)

        async with self._valkeySessionCreator() as client:
            assert taskId in self._runningTasksByTaskId, (
                "Can not cancel task %s, it is not in the running tasks"
                % taskId
            )
            taskResult = self._runningTasksByTaskId.pop(taskId)

            assert not taskResult.called, (
                "The taskResult for task %s has already been called" % taskId
            )
            taskResult.errback(
                Exception(
                    f"Task cancelled: {reason}" if reason else "Task cancelled"
                )
            )

            # Update task status
            taskKey = f"{self.taskQueueKey}:{taskId}"
            taskData = await client.get(taskKey)
            if taskData:
                task = self.deserialise(taskData)
                task["status"] = TaskState.CANCELLED.value
                task["cancelled_at"] = now.isoformat()
                task["cancellation_reason"] = reason

                taskSerialized = self.serialise(task)
                await client.set(taskKey, taskSerialized)

            # Queue cancellation result using vortex serialization
            # We queue this, so that any blocking waits will also wake up.
            error = f"Task cancelled: {reason}" if reason else "Task cancelled"
            resultTuple = PeekWorkerTaskResultTuple()
            resultTuple.taskId = taskId
            resultTuple.result = None
            resultTuple.error = error
            resultTuple.completedAt = now
            resultTuple.startedAt = None
            resultTuple.finishedAt = None
            resultTuple.dequeuedAt = None
            resultTuple.cancelledAt = now
            resultTuple.cancellationReason = reason
            resultSerialized = self.serialise(resultTuple)
            await client.lpush(self.resultQueueKey, resultSerialized)

    @convertAsyncioMethodToDeferred
    async def _removeOldTasksWrapper(self):
        oneHourAgo = datetime.now(timezone.utc) - timedelta(hours=1)

        async with self._valkeySessionCreator() as client:
            taskKeys = await client.keys(f"{self.taskQueueKey}:*")
            keysToDelete = []

            for taskKey in taskKeys:
                taskData = await client.get(taskKey)
                if not taskData:
                    continue

                task = self.deserialise(taskData)
                dequeuedAtStr = task.get("dequeued_at")

                if not dequeuedAtStr:
                    continue

                dequeuedAt = datetime.fromisoformat(dequeuedAtStr)
                if dequeuedAt < oneHourAgo:
                    taskId = task["_id"]
                    keysToDelete.extend(
                        [taskKey, f"{self.taskArgsKey}:{taskId}"]
                    )

            if keysToDelete:
                await client.delete(*keysToDelete)
                await client.delete(*keysToDelete)
                logger.info(f"Removed {len(keysToDelete)} old task keys")

    async def _cleanupCollectionsAsync(self):
        async with self._valkeySessionCreator() as client:
            # Get all keys matching our patterns
            taskKeys = await client.keys(f"{self.taskQueueKey}:*")
            taskArgsKeys = await client.keys(f"{self.taskArgsKey}:*")

            # Delete all task, task args keys and the single result queue
            allKeys = (
                taskKeys
                + taskArgsKeys
                + [self.pendingTasksKey, self.resultQueueKey]
            )
            if allKeys:
                await client.delete(*allKeys)

    def close(self):
        self.shutdown()
        if self._valkeySessionCreator:
            try:
                # Use aclose() for async connection pool
                defer.ensureDeferred(
                    self._valkeySessionCreator.disconnectAsync()
                )
            except Exception as e:
                logger.debug(f"Error disconnecting connection pool: {e}")



from datetime import datetime
from datetime import timezone
from typing import Any
from typing import NamedTuple
from typing import Optional

from twisted.internet import defer

from peek_worker_service.peek_worker_request_queue import PeekWorkerRequestQueue


class _PeekTaskResultTuple(NamedTuple):
    success: bool
    result: Optional[Any]
    error: Optional[str]
    timeoutError: Optional[bool] = False


class PeekTaskResult(defer.Deferred):
    def __init__(
        self,
        taskRequestManager: PeekWorkerRequestQueue,
        taskId: str,
        timeout: float,
    ):
        defer.Deferred.__init__(self)
        self.taskRequestManager = taskRequestManager
        self._taskId = taskId
        self._timeout = timeout
        self._startTime = datetime.now(timezone.utc)

    @property
    def taskId(self) -> str:
        return self._taskId

    @property
    def hasTimedOut(self) -> bool:
        elapsed = (datetime.now(timezone.utc) - self._startTime).total_seconds()
        return elapsed >= self._timeout



import logging
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple

import valkey.exceptions
from twisted.internet.defer import inlineCallbacks
from vortex.Payload import Payload

from peek_worker_service.peek_worker_task_state import TaskState
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskArgsTuple
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskResultTuple
from peek_worker_service.peek_worker_valkey_session import (
    PeekValkeySessionFactory,
)
from peek_worker_service.peek_worker_valkey_session import (
    convertAsyncioMethodToDeferred,
)

logger = logging.getLogger(__name__)


class PeekWorkerProcessQueue:
    def __init__(self, url: str = "valkey://localhost:6379/0"):
        self.url = url
        self._valkeySessionFactory: PeekValkeySessionFactory = None

        self.taskQueueKey = "task_queue"
        self.taskArgsKey = "task_args"
        self.resultQueueKey = "result_queue"
        self.pendingTasksKey = "pending_tasks"
        self._isAlive = True

    def serialise(self, tuple_) -> str:
        # noinspection PyProtectedMember
        return Payload(tuples=[tuple_])._toJson()

    # noinspection SpellCheckingInspection
    def deserialise(self, jsonStr: str):
        # noinspection PyProtectedMember
        return Payload()._fromJson(jsonStr).tuples[0]

    @property
    def isActive(self) -> bool:
        return self._isAlive and self._valkeySessionFactory is not None

    @inlineCallbacks
    def start(self):
        yield self._startAsync()

    @convertAsyncioMethodToDeferred
    async def _startAsync(self):
        self._valkeySessionFactory = PeekValkeySessionFactory(
            url=self.url, maxConnections=4
        )

    @inlineCallbacks
    def shutdown(self):
        self._isAlive = False
        if not self._valkeySessionFactory:
            return

        try:
            yield self._shutdownAsync()

        except Exception as e:
            logger.debug(f"Error closing Valkey session factory: {e}")

    @convertAsyncioMethodToDeferred
    async def _shutdownAsync(self):
        await self._valkeySessionFactory.disconnectAsync()

    @convertAsyncioMethodToDeferred
    async def waitForTask(
        self, workerId: str
    ) -> Tuple[Optional[Dict[str, Any]], PeekWorkerTaskArgsTuple]:
        logger.debug(f"entering waitForTask")

        async with self._valkeySessionFactory() as client:
            while self.isActive:
                # Use BRPOP to block for 5 seconds to allow graceful shutdown
                try:
                    result = await client.brpop(
                        [self.pendingTasksKey], timeout=5
                    )
                except valkey.exceptions.TimeoutError:
                    continue

                if not result:
                    continue

                # result is a tuple: (key, taskId)
                taskId = result[1]

                # Get task details
                taskKey = f"{self.taskQueueKey}:{taskId}"
                taskData = await client.get(taskKey)

                if not taskData:
                    logger.debug(f"task {taskId[:8]} not found")
                    continue

                task = self.deserialise(taskData)

                if task.get("status") != TaskState.PENDING.value:
                    logger.debug(f"task {taskId[:8]} no longer pending")
                    continue

                # Get task arguments from separate key
                taskArgsKey = f"{self.taskArgsKey}:{taskId}"
                taskArgsData = await client.get(taskArgsKey)

                if not taskArgsData:
                    logger.debug(f"task args {taskId[:8]} not found")
                    continue

                taskArgs = self.deserialise(taskArgsData)

                # Mark task as processing
                now = datetime.now(timezone.utc)
                task["status"] = TaskState.PROCESSING.value
                task["worker_id"] = workerId
                task["started_at"] = now.isoformat()
                task["dequeued_at"] = now.isoformat()

                taskSerialized = self.serialise(task)
                await client.set(taskKey, taskSerialized)

                logger.debug(f"successfully dequeued task {task['_id'][:8]}")
                return task, taskArgs

            return None, PeekWorkerTaskArgsTuple()

    @convertAsyncioMethodToDeferred
    async def queueResult(
        self, taskId: str, resultTuple: PeekWorkerTaskResultTuple
    ):
        logger.debug(
            f"Queuing result for task {taskId[:8]}, error: {resultTuple.error is not None}"
        )
        now = datetime.now(timezone.utc)

        # Update the result tuple with timing information
        resultTuple.taskId = taskId
        resultTuple.completedAt = now

        async with self._valkeySessionFactory() as client:
            # Get task info to populate timing fields and update final status
            taskKey = f"{self.taskQueueKey}:{taskId}"
            taskData = await client.get(taskKey)
            if taskData:

                def parseDateTime(dateStr):
                    return datetime.fromisoformat(dateStr) if dateStr else None

                task = self.deserialise(taskData)
                resultTuple.startedAt = parseDateTime(task.get("started_at"))
                resultTuple.dequeuedAt = parseDateTime(task.get("dequeued_at"))

                # Update task status
                task["status"] = (
                    TaskState.COMPLETED.value
                    if resultTuple.error is None
                    else TaskState.FAILED.value
                )
                task["finished_at"] = now.isoformat()
                taskSerialized = self.serialise(task)
                await client.set(taskKey, taskSerialized)

            # Push result to the single result queue using vortex serialization
            resultKey = self.resultQueueKey
            resultSerialized = self.serialise(resultTuple)
            await client.lpush(resultKey, resultSerialized)

            # Clean up task arguments since task is complete
            taskArgsKey = f"{self.taskArgsKey}:{taskId}"
            await client.delete(taskArgsKey)

    def close(self):
        self.shutdown()



import logging
import uuid
from datetime import datetime
from datetime import timezone
from typing import Callable
from typing import Dict
from typing import Optional

from twisted.internet import reactor
from twisted.internet.defer import Deferred

from peek_worker_service.peek_worker_request_queue import PeekWorkerRequestQueue
from peek_worker_service.peek_worker_task_state import TaskState
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskArgsTuple
from peek_worker_service.peek_worker_tuples import PeekWorkerTaskStartTuple


# Global task registry
_taskRegistry: Dict[str, "PeekWorkerTask"] = {}

logger = logging.getLogger(__name__)


class PeekWorkerTask:
    def __init__(self, retries: int, func: Callable):
        self.retries = retries
        self.func = func
        self.funcName = f"{func.__module__}.{func.__name__}"
        self.taskRequestManager: Optional[PeekWorkerRequestQueue] = None
        self._twistedRunning = True

        # Register the function in the global registry
        _taskRegistry[self.funcName] = self

        assert retries >= 0, "retries needs to be greater or equal to 0"

        def setStopped():
            self._twistedRunning = False

        reactor.addSystemEventTrigger("before", "shutdown", setStopped)

    def delay(self, *args, timeout: float = 1800.0, **kwargs):
        if not self._twistedRunning:
            logger.debug(
                "Not starting task %s, Reactor is stopping", self.funcName
            )

            # Return a deferred that won't fire.
            return Deferred()

        if self.taskRequestManager is None:
            raise RuntimeError(
                f"Task method {self.funcName} not initialized with taskRequestManager. "
                "Call initializeTaskMethod() first."
            )

        taskArgs = PeekWorkerTaskArgsTuple()
        taskArgs.taskMethod = self.funcName
        taskArgs.args = list(args)
        taskArgs.kwargs = kwargs

        requestArgs = PeekWorkerTaskStartTuple()
        requestArgs.taskId = str(uuid.uuid4())
        requestArgs.status = TaskState.PENDING.value
        now = datetime.now(timezone.utc)
        requestArgs.createdAt = now
        requestArgs.queuedAt = now
        requestArgs.workerId = None
        requestArgs.startedAt = None
        requestArgs.dequeuedAt = None
        requestArgs.finishedAt = None
        requestArgs.cancelledAt = None
        requestArgs.cancellationReason = None
        return self.taskRequestManager.queueTask(
            self.funcName, requestArgs, taskArgs, timeout
        )

    def __call__(self, *args, **kwargs):
        for retriesLeft in reversed(range(self.retries + 1)):
            try:
                return self.func(*args, **kwargs)
            except Exception as e:
                if not retriesLeft:
                    raise

                logger.warning(
                    "Task %s failed, retrying %s more times, %s",
                    self.funcName,
                    retriesLeft,
                    str(e),
                )

        raise Exception("This should never happen")

    def initializeTaskMethod(self, taskRequestManager: PeekWorkerRequestQueue):
        self.taskRequestManager = taskRequestManager
        return self


def peekWorkerTaskDecorator(retries: int):
    def wrapper(func: Callable) -> PeekWorkerTask:
        return PeekWorkerTask(retries, func)

    return wrapper


def initializeAllTaskMethods(
    taskRequestManager: PeekWorkerRequestQueue,
) -> Dict[str, PeekWorkerTask]:
    for funcName, func in _taskRegistry.items():
        func.initializeTaskMethod(taskRequestManager)

    return _taskRegistry



from datetime import datetime
from typing import Any
from typing import Optional

from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class PeekWorkerTaskArgsTuple(Tuple):
    __tupleType__ = "PeekWorkerTaskArgsTuple"
    taskMethod: str = TupleField()
    args: list = TupleField()
    kwargs: dict = TupleField()


@addTupleType
class PeekWorkerTaskStartTuple(Tuple):
    __tupleType__ = "PeekWorkerTaskStartTuple"
    taskId: str = TupleField()
    status: str = TupleField()
    createdAt: datetime = TupleField()
    queuedAt: datetime = TupleField()
    workerId: str = TupleField()
    startedAt: datetime = TupleField()
    dequeuedAt: datetime = TupleField()
    finishedAt: datetime = TupleField()
    cancelledAt: datetime = TupleField()
    cancellationReason: str = TupleField()


@addTupleType
class PeekWorkerTaskResultTuple(Tuple):
    __tupleType__ = "PeekWorkerTaskResultTuple"
    taskId: str = TupleField()
    result: Any = TupleField()
    error: Optional[str] = TupleField()
    completedAt: datetime = TupleField()
    startedAt: Optional[datetime] = TupleField()
    finishedAt: Optional[datetime] = TupleField()
    dequeuedAt: Optional[datetime] = TupleField()
    cancelledAt: Optional[datetime] = TupleField()
    cancellationReason: Optional[str] = TupleField()



import logging
import multiprocessing
import random
from datetime import datetime
from datetime import timezone
from typing import Any
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional

import psutil
from valkey import ValkeyError
from vortex.Tuple import TUPLE_TYPES
from vortex.Tuple import TUPLE_TYPES_BY_NAME
from vortex.Tuple import TUPLE_TYPES_BY_SHORT_NAME

from peek_worker_service.peek_worker_process import PeekWorkerProcess
from peek_worker_service.peek_worker_valkey_session import (
    PeekValkeySessionFactory,
)
from peek_worker_service.peek_worker_valkey_session import (
    convertAsyncioMethodToDeferred,
)

logger = logging.getLogger(__name__)


class WorkerConfig(NamedTuple):
    url: str
    numWorkers: int
    maxWorkerTasksCompleted: int
    maxWorkerAgeSeconds: int
    maxMemoryMb: int
    enabledPlugins: list[str]


class WorkerState:
    def __init__(
        self, process: multiprocessing.Process, worker: PeekWorkerProcess
    ):
        self.process = process
        self.worker = worker
        self.startTime = datetime.now(timezone.utc)

    @property
    def workerAgeSeconds(self) -> float:
        return (datetime.now(timezone.utc) - self.startTime).total_seconds()

    @property
    def workerMemoryMb(self) -> int:
        processInfo = psutil.Process(self.process.pid)
        return int(processInfo.memory_info().rss / 1024 / 1024)

    def shouldRecycle(self, config: WorkerConfig) -> bool:
        if not self.process.is_alive():
            return True

        if self.worker.totalTasksProcessed > (
            config.maxWorkerTasksCompleted * random.uniform(0.9, 1.1)
        ):
            return True

        if self.workerAgeSeconds > (
            config.maxWorkerAgeSeconds * random.uniform(0.9, 1.1)
        ):
            return True

        try:
            processInfo = psutil.Process(self.process.pid)
            memoryMb = processInfo.memory_info().rss / 1024 / 1024
            if memoryMb > config.maxMemoryMb:
                return True

        except (psutil.NoSuchProcess, psutil.AccessDenied):
            return True

        return False


class PeekWorkerManager:
    def __init__(self, config: WorkerConfig):
        # Set multiprocessing start method to spawn on Linux
        try:
            multiprocessing.set_start_method("spawn", force=True)
        except RuntimeError:
            # Start method already set, continue
            pass

        self.config = config
        self._workers: Dict[str, WorkerState] = {}
        self._isRunning = False
        self._shutdownRequested = False
        self._valkeySessionFactory: Optional[PeekValkeySessionFactory] = None

    async def _connectToValkeyAsync(self):
        self._valkeySessionFactory = PeekValkeySessionFactory(
            url=self.config.url, maxConnections=2
        )
        # Test connection - will raise exception if connection fails
        async with self._valkeySessionFactory() as client:
            await client.ping()
            logger.info(f"Connected to Valkey at {self.config.url}")

    async def _disconnectFromValkeyAsync(self):
        if not self._valkeySessionFactory:
            return

        await self._valkeySessionFactory.disconnectAsync()
        self._valkeySessionFactory = None
        logger.info("Disconnected from Valkey")

    async def _pingValkeyAsync(self):
        if not self._valkeySessionFactory:
            return

        try:
            async with self._valkeySessionFactory() as client:
                await client.ping()
        except ValkeyError as e:
            logger.exception(f"Valkey ping failed: {e}")

    def _createWorker(self) -> WorkerState:
        worker = PeekWorkerProcess(
            TUPLE_TYPES=TUPLE_TYPES.copy(),
            TUPLE_TYPES_BY_NAME=TUPLE_TYPES_BY_NAME.copy(),
            TUPLE_TYPES_BY_SHORT_NAME=TUPLE_TYPES_BY_SHORT_NAME.copy(),
            url=self.config.url,
            enabledPlugins=self.config.enabledPlugins,
        )
        process = multiprocessing.Process(target=worker.start)
        process.start()
        return WorkerState(process, worker)

    def _recycleWorkers(self):
        if self._shutdownRequested:
            return

        workersToRecycle = []

        for workerId, workerState in self._workers.items():
            if workerState.shouldRecycle(self.config):
                workersToRecycle.append(workerId)

        for workerId in workersToRecycle:
            logger.info(f"Recycling worker {workerId}")
            self._stopWorker(workerId)

    async def _maintainWorkerCountAsync(self):
        if self._shutdownRequested:
            return

        self._recycleWorkers()
        await self._pingValkeyAsync()

        while (
            len(self._workers) < self.config.numWorkers
            and not self._shutdownRequested
        ):
            workerState = self._createWorker()
            workerId = f"worker_{workerState.process.pid}"
            self._workers[workerId] = workerState
            logger.info(
                f"Started worker {workerId} (PID: {workerState.process.pid})"
            )

    @convertAsyncioMethodToDeferred
    async def start(self):
        assert not self._isRunning, "Workers are already running"

        logger.info(f"Starting {self.config.numWorkers} workers")

        await self._connectToValkeyAsync()
        await self._maintainWorkerCountAsync()
        self._isRunning = True

    @convertAsyncioMethodToDeferred
    async def shutdown(self):
        if not self._isRunning:
            logging.debug("Workers are not running")
            return

        self._shutdownRequested = True
        logger.info("Stopping all workers")
        workerIds = list(self._workers.keys())

        for workerId in workerIds:
            self._stopWorkerGracefully(workerId)

        await self._disconnectFromValkeyAsync()
        self._isRunning = False

    def _stopWorkerGracefully(self, workerId: str):
        if workerId not in self._workers:
            return

        workerState = self._workers[workerId]
        if not workerState.process.is_alive():
            if workerId in self._workers:
                del self._workers[workerId]
            return

        logger.info(f"Sending SIGTERM to worker {workerId}")
        workerState.process.terminate()

        # Wait up to 10 seconds for a graceful shutdown
        workerState.process.join(timeout=10)

        if workerState.process.is_alive():
            logging.warning(
                f"Worker {workerId} did not shut down gracefully, sending SIGKILL"
            )
            workerState.process.kill()
            workerState.process.join(timeout=5)

            if workerState.process.is_alive():
                logging.error(f"Worker {workerId} still alive after SIGKILL")

        if workerId in self._workers:
            del self._workers[workerId]
        logger.info(f"Worker {workerId} stopped")

    def _stopWorker(self, workerId: str):
        # Legacy method for recycling - use graceful shutdown
        self._stopWorkerGracefully(workerId)

    def getWorkerStatus(self) -> List[Dict[str, Any]]:
        status = []

        for workerId, workerState in self._workers.items():
            try:

                status.append(
                    {
                        "worker_id": workerId,
                        "pid": workerState.process.pid,
                        "is_alive": workerState.process.is_alive(),
                        "uptime_seconds": workerState.workerAgeSeconds,
                        "memory_mb": workerState.workerMemoryMb,
                    }
                )
            except (psutil.NoSuchProcess, psutil.AccessDenied):
                status.append(
                    {
                        "worker_id": workerId,
                        "pid": workerState.process.pid,
                        "is_alive": False,
                        "uptime_seconds": 0,
                        "memory_mb": 0,
                    }
                )

        return status
